"""
This script imports a SpeedTree STMAT files to Blender.

Usage:
Run this script from "File->Import" menu and then load the desired STMAT file.

https://www.speedtree.com
"""

import time
import bpy
import sys
import xml.dom.minidom as xmldom
import os.path as path
from pprint import pprint

class SpeedTreeMap:
    def __init__(self, red = 1.0, green = 1.0, blue = 1.0, file = ""):
        self.red = red
        self.green = green
        self.blue = blue
        self.file = file

class SpeedTreeMaterial:
    def __init__(self, name, twoSided = False, vertexOpacity = False, userData = ""):
        self.name = name
        self.twoSided = twoSided
        self.vertexOpacity = vertexOpacity
        self.userData = userData
        self.maps = { }
        self.shader = None
        
        
def UpdateMaterial(material, stMaterial):
    material.use_nodes = True
    material.use_backface_culling = False
    nodes = material.node_tree.nodes
    for node in nodes:
        if node.type != 'OUTPUT_MATERIAL': # skip the material output node as we'll need it later
           nodes.remove(node)
           
    links = material.node_tree.links 
    outputNode = nodes[0]
    outputNode.location = (0, 0)
    
    mainShader = nodes.new("ShaderNodeBsdfPrincipled")
    mainShader.location = (-800, -100)
    
    if ("Color" in stMaterial.maps):
        stmap = stMaterial.maps["Color"]
        if (stmap.file):
            textureNode = nodes.new("ShaderNodeTexImage")
            textureNode.location = (-1200, -100)
            textureNode.image = bpy.data.images.load(stmap.file)
            links.new(mainShader.inputs['Base Color'], textureNode.outputs['Color'])
        else:
            mainShader.inputs['Base Color'].default_value = (stmap.red, stmap.blue, stmap.green, 1)
         
    if ("Gloss" in stMaterial.maps):
        stmap = stMaterial.maps["Gloss"]
        if (stmap.file):
            textureNode = nodes.new("ShaderNodeTexImage")
            textureNode.location = (-1600, -200)
            textureNode.image = bpy.data.images.load(stmap.file)
            textureNode.image.colorspace_settings.name = 'Non-Color'
            invertNode = nodes.new("ShaderNodeInvert")
            invertNode.location = (-1150, -400)
            links.new(invertNode.inputs['Color'], textureNode.outputs['Color'])
            links.new(mainShader.inputs['Roughness'], invertNode.outputs['Color'])
        else:
            mainShader.inputs['Roughness'].default_value = 1.0 - stmap.red
            
    opacityNode = None
    if ("Opacity" in stMaterial.maps):
        stmap = stMaterial.maps["Opacity"]
        if (stmap.file):
            material.use_backface_culling = False
            material.blend_method = 'CLIP'
            material.shadow_method = 'CLIP'
            material.alpha_threshold = 0.25
            opacityNode = nodes.new("ShaderNodeTexImage")
            opacityNode.location = (-1200, 200)
            opacityNode.image = bpy.data.images.load(stmap.file)
            links.new(mainShader.inputs['Alpha'], opacityNode.outputs['Color'])
    elif (stMaterial.vertexOpacity):
        vertexColorNode = nodes.new("ShaderNodeAttribute")
        vertexColorNode.location = (-1300, 100)
        vertexColorNode.attribute_name = "blend_ao"
        separateRgbNode = nodes.new("ShaderNodeSeparateRGB")
        separateRgbNode.location = (-1100, 100)
        links.new(separateRgbNode.inputs['Image'], vertexColorNode.outputs['Color'])
        links.new(mainShader.inputs['Alpha'], separateRgbNode.outputs['R'])
        
    geometryNode = None
    normalNode = None
    if ("Normal" in stMaterial.maps):
        stmap = stMaterial.maps["Normal"]
        if (stmap.file):
            textureNode = nodes.new("ShaderNodeTexImage")
            textureNode.location = (-2000, -800)
            textureNode.image = bpy.data.images.load(stmap.file)
            textureNode.image.colorspace_settings.name = 'Non-Color'
            if (stMaterial.twoSided):
                flipBacksideNode = nodes.new("ShaderNodeRGBCurve")         
                flipBacksideNode.location = (-1700, -900)
                flipBacksideNode.mapping.curves[0].points[0].location.y = 1
                flipBacksideNode.mapping.curves[0].points[1].location.y = 0
                flipBacksideNode.mapping.curves[1].points[0].location.y = 1
                flipBacksideNode.mapping.curves[1].points[1].location.y = 0
                flipBacksideNode.mapping.update()
                flipBacksideNode.hide = True
                links.new(flipBacksideNode.inputs['Color'], textureNode.outputs['Color'])
                geometryNode = nodes.new("ShaderNodeNewGeometry")        
                geometryNode.location = (-1600, -600)
                mixBacksideNode = nodes.new("ShaderNodeMixRGB")      
                mixBacksideNode.location = (-1400, -700)
                links.new(mixBacksideNode.inputs[0], geometryNode.outputs['Backfacing'])
                links.new(mixBacksideNode.inputs[1], textureNode.outputs['Color'])
                links.new(mixBacksideNode.inputs[2], flipBacksideNode.outputs['Color'])            
                normalNode = nodes.new("ShaderNodeNormalMap")     
                normalNode.location = (-1150, -600)
                links.new(normalNode.inputs['Color'], mixBacksideNode.outputs['Color'])
                links.new(mainShader.inputs['Normal'], normalNode.outputs['Normal'])
            else:
                normalNode = nodes.new("ShaderNodeNormalMap")            
                normalNode.location = (-1150, -600)
                links.new(normalNode.inputs['Color'], textureNode.outputs['Color'])
                links.new(mainShader.inputs['Normal'], normalNode.outputs['Normal'])
          
    subsurfaceAmountNode = None      
    if ("SubsurfaceAmount" in stMaterial.maps):
        stmap = stMaterial.maps["SubsurfaceAmount"]
        if (stmap.file):
            subsurfaceAmountNode = nodes.new("ShaderNodeTexImage")   
            subsurfaceAmountNode.location = (-1600, -1000)
            subsurfaceAmountNode.image = bpy.data.images.load(stmap.file)
            subsurfaceAmountNode.image.colorspace_settings.name = 'Non-Color'
        else:
            subsurfaceAmountNode = nodes.new("ShaderNodeRGB")  
            subsurfaceAmountNode.location = (-1600, -1000)
            subsurfaceAmountNode.outputs['Color'].default_value = (stmap.red, stmap.green, stmap.blue, 1.0)
    
    subsurfaceColorNode = None      
    if ("SubsurfaceColor" in stMaterial.maps):
        stmap = stMaterial.maps["SubsurfaceColor"]
        if (stmap.file):
            subsurfaceColorNode = nodes.new("ShaderNodeTexImage") 
            subsurfaceColorNode.location = (-1300, -1200)
            subsurfaceColorNode.image = bpy.data.images.load(stmap.file)
        else:
            subsurfaceColorNode = nodes.new("ShaderNodeRGB")
            subsurfaceColorNode.location = (-1300, -1200)
            subsurfaceColorNode.outputs['Color'].default_value = (stmap.red, stmap.green, stmap.blue, 1.0)
            
    if ((subsurfaceAmountNode is None) or (subsurfaceColorNode is None)):
        # no subsurface
        if (subsurfaceAmountNode is not None):
            nodes.remove(subsurfaceAmountNode)
        if (subsurfaceColorNode is not None):
            nodes.remove(subsurfaceColorNode)
        links.new(outputNode.inputs['Surface'], mainShader.outputs['BSDF'])
    else:
        scaleSubsurfaceNode = nodes.new("ShaderNodeMath")
        scaleSubsurfaceNode.location = (-1150, -900)
        scaleSubsurfaceNode.label = "ScaleSubsurface"
        scaleSubsurfaceNode.operation = "MULTIPLY"
        scaleSubsurfaceNode.use_clamp = True
        links.new(scaleSubsurfaceNode.inputs[0], subsurfaceAmountNode.outputs['Color'])
        scaleSubsurfaceNode.inputs[1].default_value = 1.0
        subsurfaceNode = nodes.new("ShaderNodeBsdfRefraction")
        subsurfaceNode.location = (-750, -800)
        subsurfaceNode.distribution = "GGX"
        subsurfaceNode.inputs[1].default_value = 0.5
        links.new(subsurfaceNode.inputs['Color'], subsurfaceColorNode.outputs['Color'])
        subsurfaceIOR = nodes.new("ShaderNodeMixRGB") 
        if (geometryNode is None):
            geometryNode = nodes.new("ShaderNodeNewGeometry")        
            geometryNode.location = (-1600, -600)
        links.new(subsurfaceIOR.inputs[0], geometryNode.outputs['Backfacing'])
        frontIOR = nodes.new("ShaderNodeValue")
        frontIOR.outputs[0].default_value = 1.45
        links.new(subsurfaceIOR.inputs[1], frontIOR.outputs[0])
        backIOR = nodes.new("ShaderNodeValue")
        backIOR.outputs[0].default_value = 0.670
        links.new(subsurfaceIOR.inputs[2], backIOR.outputs[0])
        links.new(subsurfaceNode.inputs['IOR'], subsurfaceIOR.outputs[0])
        if (normalNode is not None):
            links.new(subsurfaceNode.inputs['Normal'], normalNode.outputs['Normal'])
        mixSubsurfaceNode = nodes.new("ShaderNodeMixShader")
        mixSubsurfaceNode.location = (-400, -500)
        links.new(mixSubsurfaceNode.inputs[0], scaleSubsurfaceNode.outputs['Value'])
        links.new(mixSubsurfaceNode.inputs[1], mainShader.outputs['BSDF'])
        links.new(mixSubsurfaceNode.inputs[2], subsurfaceNode.outputs['BSDF'])
        if (opacityNode is None):
            links.new(outputNode.inputs['Surface'], mixSubsurfaceNode.outputs['Shader'])
        else:
            finalOpacityNode = nodes.new("ShaderNodeBsdfTransparent")
            finalOpacityNode.location = (-400, -100)
            finalMixNode = nodes.new("ShaderNodeMixShader")    
            finalMixNode.location = (-200, 0)
            links.new(finalMixNode.inputs[0], opacityNode.outputs['Color'])
            links.new(finalMixNode.inputs[1], finalOpacityNode.outputs['BSDF'])
            links.new(finalMixNode.inputs[2], mixSubsurfaceNode.outputs['Shader'])
            links.new(outputNode.inputs['Surface'], finalMixNode.outputs['Shader'])
                
def load_stmat(filepath):

    t = time.time()
    
    # load stmat file
    stmatFile = path.realpath(filepath)
    stmatPath = path.dirname(stmatFile).replace("\\", "/") + "/"
    try:
        doc = xmldom.parse(stmatFile)
        root = doc.getElementsByTagName("Materials");
        if not root:
            print("[" + stmatFile + "] is not a valid stmat file")
            return {'FINISHED'}
            
        mesh = root[0].attributes["Mesh"].value
        meshFile = stmatPath + mesh
        mesh = path.splitext(mesh)[0]
        
        # load speedtree materials and create shaders
        aNewMaterials = { }
        materials = root[0].getElementsByTagName("Material")
        for material in materials:
            stMaterial = SpeedTreeMaterial(material.attributes["Name"].value,
                                            material.attributes["TwoSided"].value == "1",
                                            material.attributes["VertexOpacity"].value == "1",
                                            material.attributes["UserData"].value)
            maps = material.getElementsByTagName("Map")
            for stmap in maps:
                newmap = SpeedTreeMap()
                if ("File" in stmap.attributes):
                    newmap.file = stmatPath + stmap.attributes["File"].value
                    newmap.file = newmap.file.replace("<UDIM>", "%(UDIM)d")
                elif ("Value" in stmap.attributes):
                    newmap.red = newmap.green = newmap.blue = float(stmap.attributes["Value"].value)
                else:
                    newmap.red = float(stmap.attributes["ColorR"].value)
                    newmap.green = float(stmap.attributes["ColorG"].value)
                    newmap.blue = float(stmap.attributes["ColorB"].value)

                stMaterial.maps[stmap.attributes["Name"].value] = newmap
            aNewMaterials[stMaterial.name] = stMaterial
    except:
        #print(sys.exc_info())
        print("SpeedTree ERROR: Failed to read SpeedTree stmat file")
        return {'CANCELLED'}

    #load mesh file
    extension = path.splitext(meshFile)[1]
    try:
        if extension == ".obj":
            bpy.ops.import_scene.obj(filepath=meshFile)
        elif extension == ".abc":
            bpy.ops.wm.alembic_import(filepath=meshFile, as_background_job=False)
        elif extension == ".fbx":
            bpy.ops.import_scene.fbx(filepath=meshFile, use_custom_normals=True)
    except:
        #print(sys.exc_info())
        print("SpeedTree ERROR: Failed to load mesh file [" + meshFile + "]")
        return {'CANCELLED'}
    
    # edit materials
    try:
        # gather materials loaded
        materials = { }
        objects = bpy.context.selected_objects[:]
        for object in objects:
            for slot in object.material_slots:
                materials[slot.name] = slot.material
        
        # update them
        for materialName, material in materials.items():
            if materialName in aNewMaterials:
                UpdateMaterial(material, aNewMaterials[materialName])
            elif materialName[:-2] in aNewMaterials:
                # try removing "SG"
                UpdateMaterial(material, aNewMaterials[materialName[:-2]])
            elif materialName[:-4] in aNewMaterials:
                # try removing the additional number added
                UpdateMaterial(material, aNewMaterials[materialName[:-4]])

    except:
        #print(sys.exc_info())
        print("SpeedTree ERROR: Failed to update materials")
        return {'CANCELLED'}
    
    print('\nSuccessfully imported %r in %.3f sec' % (filepath, time.time() - t))
    return {'FINISHED'}


def load(operator, context, filepath=""):
    return load_stmat(filepath)